# Copyright Amazon Web Services and its Affiliates 

from collections import Counter
from texttable import Texttable
import mxnet as mx
import subprocess
import ctypes
import copy
import json
import os

__all__ = ['compile', 'check_compatibility', 'get_neuron_supported_list', 'experimental_set_device']

path_lib = os.path.abspath(os.path.join(os.path.dirname(__file__), "libneuron_subgraph_lib.so"))
if not os.path.exists(path_lib):
    raise ImportError("Unable to find library at '%s'" % path_lib)
mx.library.load(path_lib)

try:
    _lib = ctypes.CDLL(path_lib)
except:
    print("Unable to load the library at '%s'" % path_lib)

_NEURON_BACKEND_NAME = 'NEURON'
_NEURON_INPUT_PASS_NAME = 'addNeuronInputPass'
_NEURON_SUBG_OPTYPE = '_neuron_subgraph_op'

def _parse_inputmap(sym):
    graph = json.loads(sym.tojson())
    input_shapes_and_dtypes = []
    for n in graph['nodes']:
        if n['op'] == '_neuron_subgraph_op':
            input_map = n['attrs']['__model_inputmap__']
            input_map = input_map.split(";")
            for i in range(1, len(input_map), 3):
                shape_str = input_map[i][2:-2]
                shape = [int(sh) for sh in shape_str.split(',')]
                dtype = input_map[i+1]
                input_shapes_and_dtypes.append((shape, dtype))
            break
    return input_shapes_and_dtypes

def experimental_set_device(obj, device_id):
    """
    Set device_id attribute of all neuron_subgraph_op instances 
    in the given obj. Obj can be either mxnet symbol or gluon hybridBlock.
    This API is called to target differnt models to 

    Parameters
    ----------
    obj : MXNet symbol object or gluon HybridBlock
        MXNet json symbol or hybridBlock representing graph to compile
    device_id : integer 
        ID of the neuron device on which the model is intended to run. 
         
    Returns
    -------
    model: MXNet symbol object or gluon HybridBlock
        In case the input obj is mxnet symbol, model is a copy of obj
        that has been optimized for neuron device given by device_id.

        If the input obj is gluon.HybridBlock then the same instance is 
        optimized for neuron device given by device_id.

    """
    # Make a deep copy on the obj
    shape_map = None
    if isinstance(obj, mx.symbol.symbol.Symbol):
        model = copy.deepcopy(obj)
        shape_map = _parse_inputmap(model)
    elif isinstance(obj, mx.gluon.HybridBlock):
        model = obj
        if not model._cached_graph:
            raise RuntimeError(
                "Please first call block.hybridize() with one inference.")
        shape_map = _parse_inputmap(model._cached_graph[1])
    else:
        raise ValueError("mxnet.Symbol or gluon.HybridBlock expected, " 
            "got {}".format(type(obj)))

    if len(shape_map) == 0:
        raise ValueError("Could not parse __model_inputmap__ to generate "
                "input shape maps")

    inputs = []
    for shape, dtype in shape_map:
       inputs.append(mx.nd.zeros(shape,dtype=dtype))

    options = {"device_id": device_id}
    if isinstance(model, mx.symbol.symbol.Symbol):
        model = model.optimize_for("setNeuronDevicePass", **options, \
                clear=False)
    elif isinstance(model,mx.gluon.HybridBlock):
        model.optimize_for("setNeuronDevicePass", inputs[0], \
                *inputs[1:], **options, clear=False)
    return model

def get_neuron_supported_list():
    """
    Calls neuron-cc list-operators --framework MXNET to 
    generate the list of operators supported by neuron-cc 
    compiler. 
    Parameters
    ----------
    None

    Returns
    -------
    oplist: List of strings
        List of string representing the operators supported by neuronc-cc compiler
    """
    exec_cmd = ["neuron-cc", "list-operators", "--framework", "MXNET"]
    oplist = subprocess.check_output(' '.join(exec_cmd), shell=True)
    oplist = str(oplist, 'utf-8')
    oplist = oplist.split("\n")
    return oplist[:-1]  # Remove the last element which is 

def check_compatibility(obj):
    """
    Prints model's compatibility report w.r.t neuron-cc.

    Iterates through all the operators in the graph and lets the user know 
    which of these are supported on inferentia platform. 
    
    Input object can be mxnet symbol or HybridBlock or the path to model's 
    symbol.json. Incase if hybridBlock, the user needs to call hybridize() 
    prior to calling this method. neuron-cc needs to be installed.
    
    Parameters
    ----------
    obj : Symbol or HydridBlock object or path to symbol.json 
    
    Returns
    -------
    None
    """
    if isinstance(obj, mx.symbol.symbol.Symbol):
        sym = copy.deepcopy(obj)
        sym = json.loads(sym.tojson())
    elif isinstance(obj, mx.gluon.HybridBlock):
        if not obj._cached_graph:
            raise RuntimeError(
                "Please first call block.hybridize()")
        sym = copy.deepcopy(obj._cached_graph[1])
        sym = json.loads(sym.tojson())
    elif isinstance(obj, str):
        with open(obj) as f:
            sym = json.load(f)
    else:
        raise ValueError("mxnet.Symbol or gluon.HybridBlock or " 
                "path to symbol.json expected, got {}".format(type(obj)))

    nodes = sym["nodes"]
    allNodeTypes = Counter([ node['op'] for node in nodes if node['op'] \
                != 'null'])
    table = Texttable()
    supported_list = get_neuron_supported_list()
    allNodeProp = [('Op Type','Count','Neuron Supported ?')]
    for key in allNodeTypes:
      if key in supported_list:
        allNodeProp.append((key, allNodeTypes[key], 'Yes'))
      else:
        allNodeProp.append((key, allNodeTypes[key], 'No'))
    table.add_rows(allNodeProp)
    print(table.draw())

def _extract_compiler_options(kwargs, inputs):
    """
    Takes the compiler options provided in mxnetneuron.compile() 
    and converts it into the format that is accepted by reviewSubgraph
    and addNeuronInputPass.  
    """

    # Create compiler coptions dictionary
    compiler_options = copy.deepcopy(kwargs)
    
    # Generate a list of compiler flags
    compiler_flags, keys_to_remove = [], []
    for k, v in compiler_options.items():
        if k not in ['force_incl_node_names', 'excl_op_types', 'excl_node_names']:
            if k == "flags":
                compiler_flags.extend(v)
            else:
                if isinstance(v, int):
                    v = str(v)
                compiler_flags.append(k+' '+v)
            keys_to_remove.append(k)

    # clean up the flags and other key value pairs from compiler_options
    for k in keys_to_remove :
        del compiler_options[k] 

    compiler_options['flags'] = compiler_flags 
    compiler_options['inputNames'] = [k for k in inputs.keys()]
    
    return compiler_options

def _compile_gluon(block, inputs, **kwargs):
    """
    Compiles a mxnet.gluon block. Internally this method calls two 
    graph passes: (1) partition and compile, (2) to create new input
    node to store the neuron compiled artifacts corresponding to 
    neuron-subgraphs. 

    Parameters
    ----------
    block: mx.gluon.HybridBlock
        Gluon block representing the neural network to be compiled for
        inferentia hardware. 
    inputs : a dictionary with key/value mappings for input name to input numpy arrays 
        inputs are feature inputs and not weights
        dict key is the inputs' node name
        input_shapes and input_dtypes should contain the same list of inputs
        inputs['input_node_name'] = (dim1, dim2, .., dimN)
        All inputs must be given a shape <TODO>:Add chking.Required to compute var_ids
    **kwargs(optional)
        excl_op_types : list of op types
            List of op types that will be partitioned out of neuron
            These will be run as CPU ops
        excl_node_names : list of node names
            List of op names that will be partitioned out of neuron
            These will be run as CPU ops
        force_incl_node_names : list of node names
            List of op names that will be partitioned into neuron for 
            compilation even if they are not in neuron supported list.
        --num-neuroncores : integer
            Number of neuron-cores to be used for this model.
    Return
    ------
    block: mx.gluon.HybridBlock
        The same instance of the gluon block provided in parameters is modified
        and optimized for Neuron hardware. 

    """
    options = _extract_compiler_options(kwargs, inputs)
    
    options['dedup_subgraph'] = True
    input_values = list(inputs.values())
    
    # Graph pass to partition and compile neuron subgraphs
    block.optimize_for(input_values[0], *input_values[1:], backend=_NEURON_BACKEND_NAME, static_alloc=True,\
                       static_shape=True, backend_opts=options)
    
    # Graph pass to create input nodes to hold the neuron compiled artifacts
    block.optimize_for(input_values[0], *input_values[1:], backend=_NEURON_INPUT_PASS_NAME, \
                    static_alloc=True, static_shape=True, backend_opts=options, clear=False)
    
    return block

def _compile_symbol(sym, args, auxs, inputs, **kwargs):

    # Make deep copy of the sym, args and auxs so that we donot overwrite it.
    sym_copy = copy.deepcopy(sym)
    args_copy = copy.deepcopy(args)
    auxs_copy = copy.deepcopy(auxs)
    options = _extract_compiler_options(kwargs, inputs)

    # add the inputs to args for compilation
    for k, v in inputs.items():
        args_copy[k] = v
    
    # Graph pass to partition and compile neuron subgraphs
    sym_copy = sym_copy.optimize_for(_NEURON_BACKEND_NAME, args_copy, auxs_copy, \
                                     **options, dedup_subgraph=True)
    
    # Graph pass to create input nodes to hold the neuron compiled artifacts
    sym_copy = sym_copy.optimize_for(_NEURON_INPUT_PASS_NAME, args_copy, auxs_copy, \
                                     **options, clear=False)
    
    return (sym_copy, args_copy, auxs_copy)

def _is_neuron_compiled(sym):
    """
    Returns True if the mxnet symbol has been previously compiled
    """
    sym = json.loads(sym.tojson())

    #If any _NEURON_SUBG_OPTYPE or neuron-cc version is found then it has been compiled before
    return (0 != len([n for n in sym['nodes'] if n['op'] == _NEURON_SUBG_OPTYPE]))

def compile(obj, args=None, auxs=None, inputs=None,  **kwargs):
    """
    Compiles symbol or gluon hybridBlock to execute on Neuron

    Creates a new symbol, arg_params and aux_params
    Neuron binaries are embedded in arg_params.
    
    A previously compiled Neuron symbol can not be recompiled again
    On successful compilation the neuron-cc compiler version is inserted as a symbol attr
    The presence of a neuron-cc version attr is an indication of a prior compilation
    
    If an op is not supported by Neuron it is partitioned to the CPU for CPU execution
    Compilation are considered successful even if zero ops are partitioned to Neuron
    In that case all ops run on the CPU.
    
    When Neuron supports an op, the support can be native or emulated.
    Ops supported in emulation will also execute on CPUs
    Emulated ops on CPUs have lower performance overhead than unsupported ops on CPUs
    
    Parameters
    ----------
    obj : MXNet symbol object or gluon HybridBlock
        MXNet json symbol or hybridBlock representing graph to compile
    args (optional) : NDArray params dict. Default is None.
        args params of the model (necessary if obj is symbol)
    auxs (optional) : NDArray params dict. Default is None.
        aux params of the model (necessary if obj is symbol)
    inputs (optional) : a dictionary with key/value mappings for input name to input numpy arrays 
        inputs are feature inputs and not weights
        dict key is the inputs' node name
        input_shapes and input_dtypes should contain the same list of inputs
        inputs['input_node_name'] = (dim1, dim2, .., dimN)
        All inputs must be given a shape <TODO>:Add chking.Required to compute var_ids
    **kwargs(optional)
        excl_op_types : list of op types
            List of op types that will be partitioned out of neuron
            These will be run as CPU ops
        excl_node_names : list of node names
            List of op names that will be partitioned out of neuron
            These will be run as CPU ops
        force_incl_node_names : list of node names
            List of op names that will be partitioned into neuron for 
            compilation even if they are not in neuron supported list.
        --neuroncore-pipeline-cores : integer
            Number of neuron-cores to be used for this model.

    Returns
    -------
    Tuple of compiled model.
    In case of symbol:
        sym : json MXNet symbol
            MXNet symbol compiled for Neuron
        arg_params : NDArray params dict
            args params of the model with embedded NEFF binaries
        aux_params : NDArray params dict
            aux params of the model
    In case of HybridBlock:
        same object but optimized for inferentia.

    """
    if inputs is None:
        raise ValueError("Please provide inputs dictionary. Keys should be the " 
                "name of the input tensors and values should be mx.nd arrays of "
                "correct shape and data type. Example: \n"
                "inputs = {'data': mx.nd.ones([1,2,5,5], name='data', dtype='float16')}")

    if isinstance(obj, mx.symbol.symbol.Symbol):
        if args is None:
            raise ValueError("Missing args. Please provide args in order to compile "
                    "a mxnet symbol object. Args correspoinding to a symbol can be " 
                    "obtained by loading the graph using load_checkpoint API. Example: \n"
                    "sym, args, auxs = mx.model.load_checkpoint(PREFIX, 0)")
        if auxs is None:
            raise ValueError("Missing auxs. Please provide auxs in order to compile "
                    "a mxnet symbol object. Auxs correspoinding to a symbol can be "
                    "obtained by loading the graph using load_checkpoint API. Example: \n"
                    "sym, args, auxs = mx.model.load_checkpoint(PREFIX, 0)")
        
        # Check if the model has been compiled already
        sym = copy.deepcopy(obj) 
        if _is_neuron_compiled(sym):
            print("Symbol has already been compiled. Compilation will do "
                  "nothing and return a copy of the given model.")
            return copy.deepcopy((obj, args, auxs))

        compiled_sym, compiled_arg, compiled_aux = _compile_symbol(obj, args, auxs, inputs, **kwargs) 
        return (compiled_sym, compiled_arg, compiled_aux)

    elif isinstance(obj, mx.gluon.HybridBlock):

        # Check if the model has been compiled already
        if not obj._cached_graph:
            # Hybridize and do a inference
            input_values = list(inputs.values())
            obj.hybridize(static_alloc=True, static_shape=True)
            out = obj(input_values[0], *input_values[1:]) 
        sym = copy.deepcopy(obj._cached_graph[1])
        if _is_neuron_compiled(sym):
            print("HybridBlock has already been compiled. Compilation will do "
                  "nothing and return the same instance.")
            return (obj) 
        
        compiled_block = _compile_gluon(obj, inputs, **kwargs)
        return (compiled_block)

    else:
        raise ValueError("mxnet.Symbol or gluon.HybridBlock "
                "expected, got {}".format(type(obj)))

import os
import ctypes
lib_cx_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "libcx.so"))
ctypes.cdll.LoadLibrary(lib_cx_path)
